------------------------------------------------------------
-- ButtonManager.lua
--
-- Abin
-- 2010/10/16
------------------------------------------------------------

local CreateFrame = CreateFrame
local pairs = pairs
local hooksecurefunc = hooksecurefunc
local type = type
local strfind = strfind
local InCombatLockdown = InCombatLockdown
local UnitGetIncomingHeals = UnitGetIncomingHeals
local GetRaidTargetIndex = GetRaidTargetIndex
local SetRaidTargetIconTexture = SetRaidTargetIconTexture
local UnitInRange = UnitInRange
local UnitIsVisible = UnitIsVisible
local UnitIsDead = UnitIsDead
local UnitIsGhost = UnitIsGhost
local UnitDebuff = UnitDebuff
local UnitIsConnected = UnitIsConnected
local UnitPowerType = UnitPowerType
local UnitGUID = UnitGUID
local UnitIsTapped = UnitIsTapped
local UnitIsTappedByPlayer = UnitIsTappedByPlayer
local UnitPlayerControlled = UnitPlayerControlled
local UnitCanAttack = UnitCanAttack
local UnitReaction = UnitReaction
local UnitIsPlayer = UnitIsPlayer
local UnitIsUnit = UnitIsUnit
local strlen = string.utf8len or strlen
local strsub = string.utf8sub or strsub

local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local PowerBarColor = PowerBarColor
local GameTooltip = GameTooltip
local _G = _G

local GHOST_DEBUFF, _, GHOST_DEBUFF_ICON = GetSpellInfo(9036)
local DEATH_ICONS = { dead = "Interface\\TargetingFrame\\UI-TargetingFrame-Skull", ghost = GHOST_DEBUFF_ICON }

local _, addon = ...
local manager = {}
addon.buttonManager = manager

local Blizzard_Buttons = {} -- Stores all registered Blizzard unit buttons, accessible from no where else
local guidList = {}
local managerInitialized -- Make sure the function "manager.DoInitialize" is called only once for a game session

local buttonX, buttonY = 72, 36

------------------------------------------------------------
-- Keeps tracking size changes on normal and mini unit frames
------------------------------------------------------------

manager.GetUnitFrameSize = function()
	return buttonX, buttonY
end

manager.UpdateNonBlizzardFrameSize = function(button, data)
	if not data then
		data = Blizzard_Buttons[button]
	end

	if data and not data.blizzard then
		addon:CombatSafeCall(button, "SetSize", buttonX, data.resizeType == "mini" and  buttonY / 2 or buttonY)
	end
end

manager.OnButtonSizeChanged = function(x, y)
	local button, data
	for button, data in pairs(Blizzard_Buttons) do
		manager.UpdateNonBlizzardFrameSize(button, data)
	end
	addon.layoutManager:UpdatePartyContainerSize()
end

hooksecurefunc("SetCVar", function(cvar, value)
	if (cvar == "raidFramesWidth" or cvar == "raidFramesHeight") and type(value) == "number" and value > 2 then
		if cvar == "raidFramesWidth" then
			buttonX = value
		else
			buttonY = value
		end
		manager.OnButtonSizeChanged(buttonX, buttonY)
	end
end)

local sizeFrame = CreateFrame("Frame")
sizeFrame:RegisterEvent("VARIABLES_LOADED")
sizeFrame:SetScript("OnEvent", function(self)
	buttonX = tonumber(GetCVar("raidFramesWidth") or 72) or 72
	buttonY = tonumber(GetCVar("raidFramesHeight") or 36) or 36
	if buttonX ~= 72 or buttonY ~= 36 then
		manager.OnButtonSizeChanged(buttonX, buttonY)
	end
end)

------------------------------------------------------------
-- Functions used by this file only
------------------------------------------------------------

-- Just do a "function" type check before calling, but I do need the error messages to popup!
local function _pcall(func, ...)
	if type(func) == "function" then
		return func(...)
	end
end

local function IsBlizzardButton(self)
	return Blizzard_Buttons[self]
end

local function Button_GetAddOnData(self)
	if self == addon.sampleButton then
		return self.data
	else
		return Blizzard_Buttons[self]
	end
end

local function Button_GetRaidIconIndex(self)
	return addon.db.showRaidIcon and self.displayedUnit and GetRaidTargetIndex(self.displayedUnit) or nil
end

local function CallModulesOverlayFunc(func, overlay, ...)
	if overlay.onCreateCalled then
		addon:CallAllModulesFunc(func, overlay, ...)
	end
end

local function OnSetUnit(self)
	local data = Button_GetAddOnData(self)
	if not data then
		return
	end

	local overlay = data.overlay
	if not overlay then
		return
	end

	local unit, realUnit, inVehicle = self.displayedUnit, self.unit, (self.inVehicle and 1 or nil)
	local guid = unit and UnitGUID(unit)
	local original = overlay.guid
	local guidChanged = original ~= guid

	if guidChanged then
		if original and guidList[original] == overlay then
			guidList[original] = nil
		end

		if guid then
			guidList[guid] = overlay
		end
	end

	if guidChanged or overlay.unit ~= unit or overlay.realUnit ~= realUnit or overlay.inVehicle ~= inVehicle then
		overlay.guid, overlay.unit, overlay.realUnit, overlay.inVehicle = guid, unit, realUnit, inVehicle
		CallModulesOverlayFunc("OnUnitChange", overlay, unit, realUnit, inVehicle, guid)
	end
end

local FUNC_GROUP = {

	SetUnit = OnSetUnit,

	UpdateRoleIcon = function(self)
		local data = Button_GetAddOnData(self)
		if not data then
			return
		end
		data.roleShown = self.roleIcon:IsShown()
		manager.OnUpdateRaidIcon(self)
	end,

	UpdateHealPrediction = function(self)
		local data = Button_GetAddOnData(self)
		if not data then
			return
		end

		local texture = data.incomingHeal
		local unit = self.displayedUnit
		if not unit or UnitIsDeadOrGhost(unit) then
			texture:Hide()
			return
		end

		local width = 0
		local healthBar = self.healthBar
		local _, maxValue = healthBar:GetMinMaxValues()
		local value = healthBar:GetValue()
		local lacks = maxValue - value
		local incoming
		if lacks > 1 then
			incoming = UnitGetIncomingHeals(unit)
			if incoming and incoming > 0 then
				if incoming > lacks then
					incoming = lacks
				end
				width = incoming / maxValue * healthBar:GetWidth()
			end
		end

		texture.incoming = incoming
		if width > 1 then
			texture:SetWidth(width)
			texture:Show()
		else
			texture:Hide()
		end
	end,

	UpdateInVehicle = function(self)
		OnSetUnit(self)
		if addon.db.useVehicleColor then
			manager.UpdateHealthColor(self)
		end
	end,

	UpdateHealthColor = function(self)
		manager.UpdateHealthColor(self)
	end,

	UpdatePowerColor = function(self)
		manager.UpdatePowerColor(self)
	end,

	UpdateMaxPower = function(self)
		local powerBar = self.powerBar
		local maxPower = select(2, self.powerBar:GetMinMaxValues())
		if maxPower and maxPower > 0 then
			powerBar:SetAlpha(1)
		else
			powerBar:SetAlpha(0)
		end
	end,

	UpdateName = function(self)
		local name = self.unit and UnitName(self.unit)
		local data = Button_GetAddOnData(self)
		if not data then
			return
		end

		data.name = name
		self.name:SetText(name)
		self.name:Show()
		manager.OnNameTextSet(self)
		manager.UpdateNameColor(self)
	end,

	UpdateStatusText = function(self)
		local data = Button_GetAddOnData(self)
		if not data then
			return
		end
		local status = self.statusText:GetText()
		if data.status ~= status then
			data.status = status
			manager.UpdateNameColor(self)
			manager.UpdateHealthColor(self)
			manager.UpdateDeathIcon(self)
		end
	end,

	UpdateAuras = function(self)
		manager.UpdateDeathIcon(self)
		if addon.db.hideBuffs then
			CompactUnitFrame_HideAllBuffs(self)
		end

		if addon.db.hideDebuffs then
			CompactUnitFrame_HideAllDebuffs(self)
		end

		if addon.db.hideDispels then
			CompactUnitFrame_HideAllDispelDebuffs(self)
		end
	end,
}

local function Frame_OnEnter(self)
	local data = Button_GetAddOnData(self)
	if not data then
		return
	end

	data.highlight:Show()

	local showtooltip = addon.db.showtooltip
	if showtooltip == 0 or (showtooltip == 1 and InCombatLockdown()) then
		GameTooltip:Hide()
	end

	CallModulesOverlayFunc("OnEnter", data.overlay)
end

local function Frame_OnLeave(self)
	local data = Button_GetAddOnData(self)
	if not data then
		return
	end
	data.highlight:Hide()
	CallModulesOverlayFunc("OnLeave", data.overlay)
end

local function ArtFrame_OnEvent(self, event, unit)
	local parent = self.parent
	if event == "RAID_TARGET_UPDATE" then
		manager.OnUpdateRaidIcon(parent)
	elseif event == "UNIT_FLAGS" then
		if unit and UnitIsUnit(unit, parent.displayedUnit) then
			manager.UpdateNameColor(parent)
			manager.UpdateHealthColor(parent)
			manager.UpdateDeathIcon(parent)
		end
	else
		manager.OnUpdateRaidIcon(parent)
		manager.UpdateNameColor(parent)
		manager.UpdateDeathIcon(parent)
	end
end

local function ArtFrame_OnShow(self)
	self:SetScript("OnEvent", ArtFrame_OnEvent)
	ArtFrame_OnEvent(self)
	addon.layoutManager:CheckFrameStatus(self.parent)
	CallModulesOverlayFunc("OnShow", self.overlay)
end

local function ArtFrame_OnHide(self)
	self:SetScript("OnEvent", nil)
	addon.layoutManager:CheckFrameStatus(self.parent)
	CallModulesOverlayFunc("OnHide", self.overlay)
end

-- Improved range updates
local function Frame_UpdateInRange(self, overlay, artFrame)
	local result
	local unit = self.displayedUnit
	if unit then
		local inRange, checked = UnitInRange(unit)
		if not checked then
			inRange = UnitIsVisible(unit)
		end

		if inRange then
			self:SetAlpha(1)
			result = 1
		else
			self:SetAlpha(0.4)
		end
	end

	if result ~= artFrame.inRange then
		artFrame.inRange = result
		CallModulesOverlayFunc("OnRangeChange", overlay, result)
	end
end

-- Unit health update smoothening
local function ArtFrame_OnUpdate(self, elapsed)
	self.elapsed = (self.elapsed or 0) + elapsed
	if self.elapsed > 0.2 then
		self.elapsed = 0
		--[[
		-- I believe this is no longer needed since WOW 4.03a
		local parent = self.parent
		local unit = parent.displayedUnit
		local health = unit and UnitHealth(unit)
		if health ~= self.origiHealth then
			self.origiHealth = health
			parent.healthBar:SetValue(health or 0)
		end
		--]]
		Frame_UpdateInRange(self.parent, self.overlay, self)
	end
end

local function HealthBar_OnValueChanged(self, value)
	local data = Button_GetAddOnData(self:GetParent())
	if not data then
		return
	end
	local texture = data.incomingHeal
	if texture:IsShown() then
		local value = self:GetValue()
		local _, maxValue = self:GetMinMaxValues()
		local incoming = texture.incoming
		local overflow = value + incoming - maxValue
		if overflow > 0 then
			incoming = incoming - overflow
			texture.incoming = incoming
			if incoming > 1 then
				texture:SetWidth(incoming / maxValue * self:GetWidth())
			else
				texture:Hide()
			end
		end
	end
end

local function InitalizeAuraButtons(frames, width, point, relativePoint)
	local i, buff
	for i, buff in ipairs(frames) do
		buff:SetSize(width, width)
		buff:SetFrameLevel(5)
		if i > 1 then
			buff:ClearAllPoints()
			buff:SetPoint(point, frames[i - 1], relativePoint)
		end
	end
end

------------------------------------------------------------
-- InitializeBlizzardButton

-- This function is called only once for every Blizzard CompactRaid unit frame after
-- its caught by the addon. It is where the ugly Duckling becomes a beautiful swan!
------------------------------------------------------------
local function InitializeBlizzardButton(self, data)

	if data.initialized then
		return
	end

	data.initialized = 1

	-- Create psuedo-spacing for buttons
	self.background:ClearAllPoints()
	self.background:SetPoint("TOPLEFT")
	self.background:SetPoint("BOTTOMRIGHT", -1, 1)

	-- Make sure health/power bars has a bar texture(object) by now
	local healthBar, powerBar = self.healthBar, self.powerBar
	healthBar:SetStatusBarTexture("Interface\\RaidFrame\\Raid-Bar-Hp-Fill")
	powerBar:SetStatusBarTexture("Interface\\RaidFrame\\Raid-Bar-Hp-Fill")
	healthBar:SetFrameLevel(2)
	powerBar:SetFrameLevel(2)

	powerBar:ClearAllPoints()
	powerBar:SetPoint("TOPLEFT", healthBar, "BOTTOMLEFT")
	powerBar:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -2, 2)

	-- Vanish those stupid borders
	self.horizTopBorder:SetAlpha(0)
	self.horizBottomBorder:SetAlpha(0)
	self.vertLeftBorder:SetAlpha(0)
	self.vertRightBorder:SetAlpha(0)
	self.horizDivider:SetAlpha(0)

	-- A texture to display predicated incoming healing, the original ones are just ugly and gay
	self.myHealPredictionBar:SetAlpha(0) -- Vanish original
	self.otherHealPredictionBar:SetAlpha(0) -- Vanish original

	local incomingHeal = healthBar:CreateTexture(nil, "ARTWORK")
	data.incomingHeal = incomingHeal
	incomingHeal:SetPoint("TOPLEFT", healthBar:GetStatusBarTexture(), "TOPRIGHT")
	incomingHeal:SetPoint("BOTTOMLEFT", healthBar:GetStatusBarTexture(), "BOTTOMRIGHT")
	incomingHeal:Hide()

	healthBar:HookScript("OnValueChanged", HealthBar_OnValueChanged)

	-- A high level frame to place role icon, name text and status text, maintains OnShow/OnHide scripts, also
	-- monitors raid icon changes. This frame is internal and usually not accessible from codes out of this file.
	local artFrame = CreateFrame("Frame", nil, self)
	data.artFrame = artFrame
	artFrame.parent = self
	artFrame:SetPoint("TOPLEFT", 1, -1)
	artFrame:SetPoint("BOTTOMRIGHT", -2, 2)
	artFrame:SetFrameLevel(4)

	self.name:SetParent(artFrame)
	self.name:SetDrawLayer("ARTWORK")
	self.statusText:SetParent(artFrame)
	self.statusText:SetDrawLayer("ARTWORK")
	self.roleIcon:SetParent(artFrame)
	self.roleIcon:SetDrawLayer("ARTWORK")
	self.readyCheckIcon:SetParent(artFrame)
	self.readyCheckIcon:SetDrawLayer("OVERLAY")
	self.aggroHighlight:SetParent(artFrame)
	self.aggroHighlight:SetDrawLayer("ARTWORK")
	self.selectionHighlight:SetParent(artFrame)
	self.selectionHighlight:SetDrawLayer("OVERLAY")

	InitalizeAuraButtons(self.buffFrames, 11, "BOTTOMRIGHT", "BOTTOMLEFT")
	InitalizeAuraButtons(self.debuffFrames, 11, "BOTTOMLEFT", "BOTTOMRIGHT")
	InitalizeAuraButtons(self.dispelDebuffFrames, 12, "RIGHT", "LEFT")

	-- Death icon
	local deathIcon = artFrame:CreateTexture(self:GetName().."DeathIcon", "OVERLAY")
	data.deathIcon = deathIcon
	deathIcon:SetSize(18, 18)
	deathIcon:SetPoint("CENTER", healthBar, "CENTER")
	deathIcon:SetTexCoord(0.08, 0.92, 0.08, 0.92)

	-- For raid target icon displaying
	local raidIcon = artFrame:CreateTexture(self:GetName().."RaidTargetIcon", "ARTWORK")
	data.raidTargetIcon = raidIcon
	raidIcon:SetSize(12, 12)
	raidIcon:SetPoint("TOPLEFT")
	raidIcon:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
	raidIcon:Hide()

	-- Highlight texture
	data.highlight = artFrame:CreateTexture(nil, "BACKGROUND")
	data.highlight:SetAllPoints(artFrame)
	data.highlight:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
	data.highlight:SetBlendMode("ADD")
	data.highlight:SetVertexColor(1, 1, 1, 0.4)
	data.highlight:Hide()

	-- Create an overlay frame for modules so they can place their own stuff on it without worrying about being obstructed
	local overlay = CreateFrame("Frame", self:GetName().."_CompactRaidAddOnOverlay", self)
	data.overlay = overlay
	artFrame.overlay = overlay
	overlay:SetAllPoints(artFrame)
	overlay:SetFrameLevel(10)
end

local function Button_HookSetScript(self, script, func)
	if func and script == "OnUpdate" then
		self:SetScript("OnUpdate", nil)
		local data = Button_GetAddOnData(self)
		if data then
			Frame_UpdateInRange(self, data.overlay, data.artFrame)
		end
	end
end

local function ActivateButton(self, data)
	if data.activated then
		return
	end

	data.activated = 1

	self:HookScript("OnEnter", Frame_OnEnter)
	self:HookScript("OnLeave", Frame_OnLeave)

	local artFrame, overlay = data.artFrame, data.overlay
	artFrame:RegisterEvent("RAID_TARGET_UPDATE")
	artFrame:RegisterEvent("UNIT_FLAGS")
	artFrame:SetScript("OnUpdate", ArtFrame_OnUpdate)
	artFrame:SetScript("OnShow", ArtFrame_OnShow)
	artFrame:SetScript("OnHide", ArtFrame_OnHide)

	addon:CallAllModulesFunc("OnCreate", overlay)
	overlay.onCreateCalled = 1
	manager.CallAllGroupFuncs(self)
	manager.ModifyButtonLayout(self)

	-- Remove Blizzard's OnUpdate script which calles every single FPS does nothing other than updating unit in-range
	-- status, and consumes a truck load of CPU resource stupidly. Range is never a critical issue that worth check every
	-- FPS, I have reduced the checking interval down to 0.2 seconds.
	-- Besides, alpha 0.55 is not translucent enough for most players to distinguish an out-ranged unit among others,
	-- 0.4 looks like a better one.
	self:SetScript("OnUpdate", nil)
	hooksecurefunc(self, "SetScript", Button_HookSetScript)

	self:HookScript("OnSizeChanged", manager.OnNameTextSet)

	-- Triggers ArtFrame_OnShow if the unit frame is already visible (usually not, though)
	if artFrame:IsVisible() then
		ArtFrame_OnShow(artFrame)
	end
end

local function RegisterButton(self, blizzard, resizeType)
	if type(self) == "string" then
		self = _G[self]
	end

	if type(self) == "table" and self.healthBar and self.powerBar and self.aggroHighlight and self.horizTopBorder and self ~= addon.sampleButton then
		local data = Blizzard_Buttons[self]
		if data then
			return 1 -- Already registered
		end

		-- We create a data table for every unit frame to store our own stuff
		data = {}
		Blizzard_Buttons[self] = data
		data.blizzard, data.resizeType = blizzard, resizeType
		InitializeBlizzardButton(self, data)

		if managerInitialized then
			ActivateButton(self, data)
		end

		return self, resizeType
	end
end

local RESIZE_TYPES = { normal = 1, mini = 1 }

manager.RegisterFrame = function(frame, resizeType)
	if type(resizeType) == "string" then
		resizeType = strlower(strtrim(resizeType))
	end

	if not RESIZE_TYPES[resizeType] then
		resizeType = nil
	end

	return RegisterButton(frame, nil, resizeType)
end

-- Scan for all existing Blizzard buttons
local function ScanBlizzardButtons()
	local i, j, name, button
	for i = 1, 5 do
		RegisterButton("CompactPartyFrameMember"..i, 1)
		RegisterButton("CompactPartyFramePet"..i, 1)
	end

	for i = 1, 80 do
		RegisterButton("CompactRaidFrame"..i, 1)
	end

	for i = 1, 8 do
		for j = 1, 5 do
			RegisterButton("CompactRaidGroup"..i.."Member"..j, 1)
		end
	end
end

------------------------------------------------------------
-- Functions may used by other files
------------------------------------------------------------

-- Enumerate all created overlay frames, your callback function returning non-nil value will abort the process
manager.EnumFrames = function(object, func, ...)
	local enumType, memberFunc
	if type(object) == "function" then
		enumType = 1 -- Direct function call, object(frame, func, ...)
	elseif type(func) == "string" then
		if not object then
			local testData = Blizzard_Buttons[CompactPartyFrameMember1]
			if not testData then
				return 0 -- if CompactPartyFrameMember1 isn't there, it's too early to enumerate
			end
			memberFunc = testData.overlay[func]
			if type(memberFunc) == "function" then
				enumType = 2 -- Call frame method, frame:func(...)
			end

		elseif type(object) == "table" then
			memberFunc = object[func]
			if type(memberFunc) == "function" then
				enumType = 3 -- Call the object's member method, object:func(frame, ...)
			end
		end
	end

	local count = 0
	local data, abort, frameMethod
	for _, data in pairs(Blizzard_Buttons) do
		count = count + 1
		local frame = data.overlay
		if enumType == 1 then
			abort = object(frame, func, ...)
		elseif enumType == 2 then
			abort = memberFunc(frame, ...)
		elseif enumType == 3 then
			abort = memberFunc(object, frame, ...)
		end

		if abort then
			break
		end
	end
	return count, abort
end

manager.GetOverlay = function(self)
	local data = Blizzard_Buttons[self]
	return data and data.overlay
end

manager.GetOverlayByGUID = function(guid)
	return guidList[guid]
end

manager.CallGroupFunc = function(func, ...)
	_pcall(FUNC_GROUP[func], ...)
end

manager.CallAllGroupFuncs = function(...)
	local func
	for _, func in pairs(FUNC_GROUP) do
		func(...)
	end
end

manager.CallAllButtonFunc = function(func, ...)
	if type(func) == "function" then
		local button
		for button in pairs(Blizzard_Buttons) do
			func(button, ...)
		end
	end
end

-- Set texture for health bar and mana bar together, along with their background textures, oh, not to forget the incoming-healing predicate
manager.SetBothBarTexture = function(self, texture)
	local data = Button_GetAddOnData(self)
	if not data then
		return
	end

	local healthBar, powerBar = self.healthBar, self.powerBar
	if not texture then
		texture = ""
	end

	healthBar:SetStatusBarTexture(texture, "ARTWORK")
	healthBar.background:SetTexture(texture)
	powerBar:SetStatusBarTexture(texture, "ARTWORK")
	powerBar.background:SetTexture(texture)
	data.incomingHeal:SetTexture(texture)
end

manager.OnStatusBarTextureSet = function(self)
	local texture = self.healthBar:GetStatusBarTexture()
	manager.SetBothBarTexture(self, texture and texture:GetTexture())
end

local function SetStatusBarColor(bar, r, g, b)
	local dr, dg, db = r * 0.25, g * 0.25, b * 0.25
	if addon.db.invertColor then
		bar.background:SetVertexColor(dr, dg, db)
		bar:SetStatusBarColor(r, g, b)
	else
		bar.background:SetVertexColor(r, g, b)
		bar:SetStatusBarColor(dr, dg, db)
	end
end

local function GetPlayerClassColor(unit, r, g, b)
	local color = RAID_CLASS_COLORS[select(2, UnitClass(unit))]
	if color then
		return color.r, color.g, color.b
	else
		return r, g, b
	end
end

local function GetNPCColor(unit, ignoreTap)
	if UnitIsDead(unit) or (not ignoreTap and UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit)) then
		return 0.5, 0.5, 0.5
	elseif UnitPlayerControlled(unit) then
		if UnitCanAttack(unit, "player") then
			return 1, 0, 0
		elseif UnitCanAttack("player", unit) then
			return 1, 1, 0
		else
			return 0.5, 0.5, 1 -- friendly pet
		end
	else
		local reaction = UnitReaction(unit, "player")
		if not reaction then
			return 0.5, 0.5, 1
		elseif reaction < 4 then
			return 1, 0, 0
		elseif reaction > 4 then
			return 0, 1, 0
		else
			return 1, 1, 0
		end
	end
end

manager.UpdateHealthColor = function(self)
	local data = Button_GetAddOnData(self)
	if not data then
		return
	end

	local unit = self.unit
	if not unit then
		return
	end

	local r, g, b
	if not UnitIsConnected(unit) then
		r, g, b = 0.5, 0.5, 0.5
	elseif self.inVehicle and addon.db.useVehicleColor then
		r, g, b = addon.db.vehicler or 0, addon.db.vehicleg or 1, addon.db.vehicleb or 0
	elseif addon.db.healthClass == 1 then
		if UnitIsPlayer(unit) then
			r, g, b = GetPlayerClassColor(unit, 0, 1, 0)
			if addon.db.nameClass == 1 then
				self.name:SetTextColor(r, g, b)
			end
		else
			r, g, b = GetNPCColor(unit, 1)
		end
	else
		r, g, b = addon.db.healthr or 0, addon.db.healthg or 1, addon.db.healthb or 0
	end

	SetStatusBarColor(self.healthBar, r, g, b)
	data.incomingHeal:SetVertexColor(r * 0.5, g * 0.5, b * 0.5)
end

manager.UpdatePowerColor = function(self)
	local unit = self.displayedUnit
	if not unit or not UnitExists(unit) then
		return
	end

	local r, g, b
	if not UnitIsConnected(unit) then
		--Color it gray
		r, g, b = 0.5, 0.5, 0.5
	else
		local powerType, powerToken, altR, altG, altB = UnitPowerType(unit)
		local color = PowerBarColor[powerToken] or PowerBarColor[powerType]
		if color then
			r, g, b = color.r, color.g, color.b
		elseif altR then
			r, g, b = altR, altG, altB
		else
			r, g, b = 0, 0, 1
		end
	end

	SetStatusBarColor(self.powerBar, r, g, b)
end

manager.UpdateNameColor = function(self)
	local unit = self.unit
	if not unit or not UnitExists(unit) then
		return
	end

	local r, g, b
	if UnitIsPlayer(unit) then
		if UnitIsDead(unit) then
			r, g, b = 1, 0, 0
		else
			if addon.db.nameClass == 1 then
				r, g, b = GetPlayerClassColor(unit, 1, 1, 1)
			else
				r, g, b = addon.db.namer, addon.db.nameg, addon.db.nameb
			end
		end
	else
		r, g, b = GetNPCColor(unit)
	end

	self.name:SetTextColor(r or 1, g or 1, b or 1)
end

manager.UpdateDeathIcon = function(self)
	local data = Button_GetAddOnData(self)
	if not data then
		return
	end

	local unit = self.displayedUnit
	if not unit or not UnitExists(unit) then
		return
	end

	local flag
	if UnitIsPlayer(unit) or UnitPlayerControlled(unit) then
		if UnitIsDead(unit) then
			flag = "dead"
		elseif UnitDebuff(unit, GHOST_DEBUFF) then
			flag = "ghost"
		end
	end

	local deathIcon = data.deathIcon
	if deathIcon.flag ~= flag then
		deathIcon.flag = flag
		deathIcon:SetTexture(DEATH_ICONS[flag])
		CallModulesOverlayFunc("OnDeathFlagChange", data.overlay, flag)
	end
end

manager.OnNameTextHeight = function(self)
	local height = addon.db.nameHeight or 12
	self.name:SetFont(STANDARD_TEXT_FONT, height)
	self.statusText:SetFont(STANDARD_TEXT_FONT, height * 0.84)
	manager.OnNameTextSet(self)
end

manager.OnNameTextOffset = function(self)
	self.name:ClearAllPoints()
	self.name:SetPoint("CENTER", self.healthBar, "CENTER", addon.db.nameXOffset, addon.db.nameYOffset)
end

manager.OnNameTextInit = function(self)
	self.name:SetJustifyH("CENTER")
	self.statusText:ClearAllPoints()
	self.statusText:SetPoint("TOP", self.name, "BOTTOM", 0, 1)
	self.statusText:SetJustifyH("CENTER")
	manager.OnNameTextHeight(self)
	manager.OnNameTextOffset(self)
end

manager.OnNameTextSet = function(self)
	local data = Button_GetAddOnData(self)
	if not data then
		return
	end

	local name = self.name
	local text = data.name or name:GetText()
	if not text then
		return
	end

	local buttonWidth = self:GetWidth()
	if buttonWidth < 1 then
		return
	end

	local widthLimit = addon.db.nameWidthLimit
	if type(widthLimit) == "number" then
		widthLimit = widthLimit / 100
	else
		widthLimit = 0.75
	end

	local i
	for i = strlen(text), 0, -1 do
		name:SetText(strsub(text, 1, i))
		if name:GetWidth() / buttonWidth <= widthLimit then
			break
		end
	end
end

manager.UpdatePowerBarHeight = function(self)
	local data = Button_GetAddOnData(self)
	if not data then
		return
	end

	if data.resizeType ~= "mini" then
		self.healthBar:SetPoint("BOTTOMRIGHT", -2, 2 + (addon.db.powerBarHeight or 5))
		self.powerBar:Show()
	end
end

manager.OnUpdateRaidIcon = function(self)
	local data = Button_GetAddOnData(self)
	if not data then
		return
	end

	local roleIcon = self.roleIcon
	local raidIcon = data.raidTargetIcon
	local index = Button_GetRaidIconIndex(self)
	local hasRaidIcon, hasRoleIcon

	if index then
		SetRaidTargetIconTexture(raidIcon, index)
		raidIcon:Show()
		roleIcon:Hide()
		hasRaidIcon = 1
	else
		raidIcon:Hide()
		if data.roleShown then
			roleIcon:Show()
			index = 1
			hasRoleIcon = 1
		end
	end

	raidIcon:SetWidth(index and 12 or 1)

	if data.hasRaidIcon ~= hasRaidIcon or data.hasRoleIcon ~= hasRoleIcon then
		data.hasRaidIcon, data.hasRoleIcon = hasRaidIcon, hasRoleIcon
		CallModulesOverlayFunc("OnIconsChange", data.overlay, hasRaidIcon, hasRoleIcon)
	end
end

manager.SetBuffFrameScalePoint = function(self, key)
	if type(key) ~= "string" then
		return
	end

	local list = self[key.."Frames"]
	if not list then
		return
	end

	local buff = list[1]
	if not buff then
		return
	end

	local scale = addon.db[key.."Scale"]
	if type(scale) == "number" and scale > 10 and scale < 500 then
		scale = scale / 100
		local button
		for _, button in ipairs(list) do
			button:SetScale(scale)
		end
	end

	local x, y = addon.db[key.."XOffset"], addon.db[key.."YOffset"]
	if type(x) ~= "number" then
		x = 0
	end

	if type(y) ~= "number" then
		y = 0
	end

	buff:ClearAllPoints()
	if key == "buff" then
		buff:SetPoint("BOTTOMRIGHT", self.healthBar, "BOTTOMRIGHT", -1 + x, 1 + y)
	elseif key == "debuff" then
		buff:SetPoint("BOTTOMLEFT", self.healthBar, "BOTTOMLEFT", 1 + x, 1 + y)
	elseif key == "dispelDebuff" then
		buff:SetPoint("TOPRIGHT", self.healthBar, "TOPRIGHT", -1 + x, -1 + y)
	end
end

manager.ModifyButtonLayout = function(self)
	local data = IsBlizzardButton(self)
	if not data then
		return
	end

	self.background:SetTexture(0, 0, 0, 1)
	self.background:SetTexCoord(0, 1, 0, 1)
	self.roleIcon:ClearAllPoints()
	self.roleIcon:SetPoint("TOPLEFT", data.raidTargetIcon, "TOPLEFT")
	self.powerBar:ClearAllPoints()
	self.powerBar:SetPoint("TOPLEFT", self.healthBar, "BOTTOMLEFT")
	self.powerBar:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -2, 2)
	self.aggroHighlight:ClearAllPoints()
	self.aggroHighlight:SetAllPoints(self.background)
	self.selectionHighlight:ClearAllPoints()
	self.selectionHighlight:SetAllPoints(self.background)
	manager.OnNameTextInit(self)
	manager.UpdatePowerBarHeight(self)
	manager.OnStatusBarTextureSet(self)
	Frame_UpdateInRange(self, data.overlay, data.artFrame)

	manager.SetBuffFrameScalePoint(self, "buff")
	manager.SetBuffFrameScalePoint(self, "debuff")
	manager.SetBuffFrameScalePoint(self, "dispelDebuff")
end

manager.DoInitialize = function()

	if managerInitialized then
		return
	end

	------------------------------------------------------------
	-- Scan for already created Blizzard buttons
	------------------------------------------------------------

	ScanBlizzardButtons()

	------------------------------------------------------------
	-- Initialize all registered unit frames
	------------------------------------------------------------

	local button, data
	for button, data in pairs(Blizzard_Buttons) do
		ActivateButton(button, data)
	end

	------------------------------------------------------------
	-- Install hooks for unit frames those will be created later
	------------------------------------------------------------

	-- Catch new Blizzard button creations
	hooksecurefunc("CompactUnitFrame_OnLoad", function(self)
		if Blizzard_Buttons[self] or type(self) ~= "table" or type(self.GetName) ~= "function" then
			return
		end

		local name = self:GetName()
		if type(name) == "string" and (strfind(name, "^CompactRaidFrame(%d+)$") or strfind(name, "^CompactRaidGroup(%d+)Member(%d+)$") or strfind(name, "^CompactPartyFrameMember(%d+)$") or strfind(name, "CompactPartyFramePet(%d+)$")) then
			RegisterButton(self, 1)
		end
	end)

	hooksecurefunc("CompactUnitFrame_UpdateAll", function(self)
		local data = IsBlizzardButton(self)
		if data then
			manager.CallAllGroupFuncs(self, data.overlay)
			Frame_UpdateInRange(self, data.overlay, data.artFrame)
		end
	end)

	local funcName
	for funcName in pairs(FUNC_GROUP) do
		hooksecurefunc("CompactUnitFrame_"..funcName, function(self)
			local data = IsBlizzardButton(self)
			if data then
				FUNC_GROUP[funcName](self, data.overlay)
			end
		end)
	end

	-- These functions are where they possibly change health/mana bar textures, anchors, text points etc
	hooksecurefunc("CompactUnitFrame_SetUpFrame", manager.ModifyButtonLayout) -- added in 0.2beta
	hooksecurefunc("DefaultCompactUnitFrameSetup", manager.ModifyButtonLayout)
	hooksecurefunc("DefaultCompactMiniFrameSetup", manager.ModifyButtonLayout)

	managerInitialized = 1
end